@@ -1,3 +1,5 @@
+use std::cmp;
+
 use crate::ansi::AnsiStyle;
 use crate::ansi::Color::Default;
 use crate::ansi::Color::Green;
@@ -20,8 +22,7 @@ pub(crate) struct StyledToken {
 }
 
 #[derive(Debug, PartialEq, Eq)]
-pub(crate) struct LineStyle<'a> {
-    prefix: &'a str,
+pub(crate) struct LineStyle {
     prefix_style: AnsiStyle,
     plain_style: AnsiStyle,
     highlighted_style: AnsiStyle,
@@ -29,7 +30,6 @@ pub(crate) struct LineStyle<'a> {
 
 pub(crate) const LINE_STYLE_OLD: LineStyle = {
     LineStyle {
-        prefix: "-",
         prefix_style: AnsiStyle {
             inverse: false,
             weight: Weight::Normal,
@@ -50,7 +50,6 @@ pub(crate) const LINE_STYLE_OLD: LineStyle = {
 
 pub(crate) const LINE_STYLE_OLD_FAINT: LineStyle = {
     LineStyle {
-        prefix: "-",
         prefix_style: AnsiStyle {
             inverse: false,
             weight: Weight::Faint,
@@ -71,7 +70,6 @@ pub(crate) const LINE_STYLE_OLD_FAINT: LineStyle = {
 
 pub(crate) const LINE_STYLE_NEW: LineStyle = {
     LineStyle {
-        prefix: "+",
         prefix_style: AnsiStyle {
             inverse: false,
             weight: Weight::Normal,
@@ -92,7 +90,6 @@ pub(crate) const LINE_STYLE_NEW: LineStyle = {
 
 pub(crate) const LINE_STYLE_ADDS_ONLY: LineStyle = {
     LineStyle {
-        prefix: "+",
         prefix_style: AnsiStyle {
             inverse: false,
             weight: Weight::Faint,
@@ -113,7 +110,6 @@ pub(crate) const LINE_STYLE_ADDS_ONLY: LineStyle = {
 
 pub(crate) const LINE_STYLE_OLD_FILENAME: LineStyle = {
     LineStyle {
-        prefix: "--- ",
         prefix_style: AnsiStyle {
             inverse: false,
             weight: Weight::Bold,
@@ -134,7 +130,6 @@ pub(crate) const LINE_STYLE_OLD_FILENAME: LineStyle = {
 
 pub(crate) const LINE_STYLE_NEW_FILENAME: LineStyle = {
     LineStyle {
-        prefix: "+++ ",
         prefix_style: AnsiStyle {
             inverse: false,
             weight: Weight::Bold,
@@ -153,50 +148,6 @@ pub(crate) const LINE_STYLE_NEW_FILENAME: LineStyle = {
     }
 };
 
-/// Like `LINE_STYLE_OLD` but without any prefix
-pub(crate) const LINE_STYLE_CONFLICT_C1: LineStyle = {
-    LineStyle {
-        prefix: "",
-        prefix_style: AnsiStyle {
-            inverse: false,
-            weight: Weight::Normal,
-            color: Red,
-        },
-        plain_style: AnsiStyle {
-            inverse: false,
-            weight: Weight::Normal,
-            color: Red,
-        },
-        highlighted_style: AnsiStyle {
-            inverse: true,
-            weight: Weight::Normal,
-            color: Red,
-        },
-    }
-};
-
-/// Like `LINE_STYLE_NEW` but without any prefix
-pub(crate) const LINE_STYLE_CONFLICT_C2: LineStyle = {
-    LineStyle {
-        prefix: "",
-        prefix_style: AnsiStyle {
-            inverse: false,
-            weight: Weight::Normal,
-            color: Green,
-        },
-        plain_style: AnsiStyle {
-            inverse: false,
-            weight: Weight::Normal,
-            color: Green,
-        },
-        highlighted_style: AnsiStyle {
-            inverse: true,
-            weight: Weight::Normal,
-            color: Green,
-        },
-    }
-};
-
 impl StyledToken {
     pub fn new(token: String, style: Style) -> StyledToken {
         return StyledToken { token, style };
@@ -216,7 +167,7 @@ impl StyledToken {
 }
 
 #[must_use]
-fn render_row(line_style: &LineStyle, row: &[StyledToken]) -> String {
+fn render_row(line_style: &LineStyle, prefix: &str, row: &[StyledToken]) -> String {
     let mut rendered = String::new();
 
     let mut current_style = ANSI_STYLE_NORMAL;
@@ -224,7 +175,7 @@ fn render_row(line_style: &LineStyle, row: &[StyledToken]) -> String {
     // Render prefix
     rendered.push_str(&line_style.prefix_style.from(&current_style));
     current_style = line_style.prefix_style;
-    rendered.push_str(line_style.prefix);
+    rendered.push_str(prefix);
 
     // Render tokens
     for token in row {
@@ -256,13 +207,13 @@ fn render_row(line_style: &LineStyle, row: &[StyledToken]) -> String {
 
 /// Render all the tokens into a (most of the time multiline) string
 #[must_use]
-pub fn render(line_style: &LineStyle, tokens: Vec<StyledToken>) -> String {
+pub fn render(line_style: &LineStyle, prefix: &str, tokens: &Vec<StyledToken>) -> String {
     let mut rendered = String::new();
 
     let mut current_row_start = 0;
     for (i, token) in tokens.iter().enumerate() {
         if token.token == "\n" {
-            let rendered_row = &render_row(line_style, &tokens[current_row_start..i]);
+            let rendered_row = &render_row(line_style, prefix, &tokens[current_row_start..i]);
             rendered.push_str(rendered_row);
             rendered.push('\n');
             current_row_start = i + 1;
@@ -271,7 +222,7 @@ pub fn render(line_style: &LineStyle, tokens: Vec<StyledToken>) -> String {
     }
 
     if current_row_start < tokens.len() {
-        let rendered_row = &render_row(line_style, &tokens[current_row_start..]);
+        let rendered_row = &render_row(line_style, prefix, &tokens[current_row_start..]);
         rendered.push_str(rendered_row);
     }
 
@@ -362,6 +313,38 @@ pub fn highlight_nonleading_tabs(tokens: &mut [StyledToken]) {
     }
 }
 
+pub(crate) fn align_tabs(old: &mut [StyledToken], new: &mut [StyledToken]) {
+    let old_tab_index_token = old.iter().position(|token| token.token == "\t");
+    if old_tab_index_token.is_none() {
+        return;
+    }
+    let old_tab_index_token = old_tab_index_token.unwrap();
+    let old_tab_index_char = old
+        .iter()
+        .take(old_tab_index_token)
+        .map(|token| token.token.chars().count())
+        .sum::<usize>();
+
+    let new_tab_index_token = new.iter().position(|token| token.token == "\t");
+    if new_tab_index_token.is_none() {
+        return;
+    }
+    let new_tab_index_token = new_tab_index_token.unwrap();
+    let new_tab_index_char = new
+        .iter()
+        .take(new_tab_index_token)
+        .map(|token| token.token.chars().count())
+        .sum::<usize>();
+
+    let old_spaces =
+        " ".repeat(2 + cmp::max(old_tab_index_char, new_tab_index_char) - old_tab_index_char);
+    let new_spaces =
+        " ".repeat(2 + cmp::max(old_tab_index_char, new_tab_index_char) - new_tab_index_char);
+
+    old[old_tab_index_token].token = old_spaces;
+    new[new_tab_index_token].token = new_spaces;
+}
+
 /// Highlight single space between two highlighted tokens
 pub fn bridge_consecutive_highlighted_tokens(tokens: &mut [StyledToken]) {
     enum FoundState {
@@ -525,7 +508,8 @@ mod tests {
     fn test_basic() {
         let rendered = render(
             &LINE_STYLE_NEW,
-            vec![
+            "+",
+            &vec![
                 StyledToken {
                     token: "hej".to_string(),
                     style: Style::Plain,
@@ -580,7 +564,8 @@ mod tests {
         // It shouldn't be highlighted, just added ones should
         let actual = render(
             &LINE_STYLE_OLD,
-            vec![StyledToken::new(" ".to_string(), Style::Plain)],
+            "-",
+            &vec![StyledToken::new(" ".to_string(), Style::Plain)],
         );
 
         assert_eq!(actual, format!("{OLD}- {NORMAL}"));
@@ -643,7 +628,8 @@ mod tests {
         // It shouldn't be highlighted, just added ones should
         let actual = render(
             &LINE_STYLE_OLD,
-            vec![
+            "-",
+            &vec![
                 StyledToken::new("x".to_string(), Style::Plain),
                 StyledToken::new("\t".to_string(), Style::Plain),
             ],
